#ifndef IMAGE_HH
#define IMAGE_HH

#include <cstdlib>
#include "image_aux.hh"

//! A namespace to prevent global namespace pollution
namespace diptools
{

  //! The primary image class
  class Image
  {
  protected:
    /*! The image data is to be stored as a matrix or 2D array.
     *  The array should be dynamically allocated during runtime.
     *  The array will store image intensity values. For now we 
     *  are working with grayscale images so Image_data_t is just 
     *  another name for double. The intensity values should range
     *  from 0.0 to 1.0.
     */
    Image_data_t** data;

    //! Width of the image
    unsigned int width;

    //! Height of the image
    unsigned int height;

    //! Bit depth of the image
    unsigned int depth;

    //! Image type. For now set this to DIPTOOLS_GRAYSCALE_IMAGE always.
    Image_type_t type;

  public:

    //---------------------------------------------------------------------
    // Constructors and Destructors
    //---------------------------------------------------------------------

    /*! Default Dummy Constructor
     *
     * Does not do anything. Do not implement this again.
     */
    Image() { data = NULL; }

    /*! Constructor
     *
     *  This creates a blank 8-bit image of size _width x _height of type _type. All the intensity
     *  values are initialized to 0.0
     */
    Image(const unsigned int _width, const unsigned int _height, const unsigned int _depth = 8, const Image_type_t _type = DIPTOOLS_GRAYSCALE_IMAGE);
    /*! Copy Constructor
     *
     *  This copies the data of image given as the parameter to form a new
     *  image. Be sure to copy the image data array and not just the pointer
     */
    Image(const Image& _img);

    /*! Destructor
     *
     * Free the image data array inside the this.
     */
    ~Image();

    //----------------------------------------------------------------------
    // Overloaded Operators
    //----------------------------------------------------------------------

    /*! Overloaded assignment operator
     *  Basically functions similarly to the copy constructor
     */
    Image& operator= ( const Image& _img );

    //----------------------------------------------------------------------
    // Get Methods
    // Do not implement these. Implemented inline.
    //----------------------------------------------------------------------

    unsigned int get_width(void) const { return width; }
    unsigned int get_height(void) const { return height; }
    unsigned int get_depth(void) const { return depth; }
    Image_type_t get_type(void) const { return type; }
    
    //----------------------------------------------------------------------
    // Read/Write Methods
    //----------------------------------------------------------------------

    /*!
     * This will open the PGM image 'pgm_image_filename' and read its contents into
     * the data array. Any previous data in the array will be wiped off. The array 
     * if already exiting should be deallocated and reallocated.
     *
     * The function should return the following:
     * 1 - If it is successful in reading the image.
     * -1 - If it cannot open/access the file.
     * -2 - If it cannot allocate memory.
     * -3 - If it cannot understand the format of the image.
     */
    int read_pgm(const char* pgm_image_filename);
    
    /*! 
     * This will write the data array as a PGM image in the file named 'pgm_image_filename'
     *
     * The function should return the following:
     * 1 - If it is successful in writing the image.
     * -1 - If it cannot open/access the file.
     */
    int write_pgm(const char* pgm_image_filename);

    /*****************************************************************************************
     * All functions will modify the existing image data i.e. the methods will
     * modify the data array of the Image object it is called with.
     *****************************************************************************************/
     

    //-------------------------------------------------------------------------
    // Transfer functions
    //-------------------------------------------------------------------------

    //! Image invert
    void invert(void);

    /*! Contrast Stretch function
     *  r1, r2 are the intensity range values in the existing image.
     *  s1, s2 are the intensity range values in the contrast stretched image.
     */
    void contrast_stretch(double r1, double r2, double s1, double s2);

    /*! Gamma Correction
     *  Here c and gamma define a transfer function :
     *                                     s = c * (r ^ gamma)           
     */  
    void gamma_correct(double c, double gamma);

    /*! Dynamic Range Compression
     * Here c defines a transfer function :
     *                                s = c * (log(1 + r))
     */
    void range_compress(double c);
 
    //-------------------------------------------------------------------------
    // Histogram Operations
    //-------------------------------------------------------------------------

    //! Returns the histogram of the image
    histogram_t get_histogram(void);

    //! Global Histogram Equilization 
    void equalize(void);

    //! Local Histogram Equilization 
    void local_equalize(unsigned int mask_size);

    //! Histogram Specification
    //! This will actually perform histgram specification from the given histogram.
    void set_histogram(const histogram_t& histogram);

    //--------------------------------------------------------------------------
    // Smoothening Filters
    //--------------------------------------------------------------------------

    //! Neighbourhood Averaging filter
    void neighbourhood_average(const mask_t &_mask);

    //! Median Filter
    void median(const mask_t &_mask);

    //--------------------------------------------------------------------------
    // Sharpening Filters
    //--------------------------------------------------------------------------

    //! Sharpening filter
    void sharpen(const mask_t &_mask);

    //! HighBoost Filter
    void high_boost(const mask_t &_mask);

    //--------------------------------------------------------------------------
    // Haltoning
    //--------------------------------------------------------------------------

    //! Halftoning function - converts the image to a halftone image
    //! dither_matrix is the pseudo-random dither matrix
    void halftone(const mask_t &dither_matrix);

    //--------------------------------------------------------------------------
    // Low Level Pixel Access
    //--------------------------------------------------------------------------

    //! Get the pixel intensity at (i, j)
    Image_data_t get_pixel(const unsigned int i, const unsigned int j);

    //! Set the pixel intensity at (i, j)
    void set_pixel(const unsigned int i, const unsigned int j, const Image_data_t &img_t);

    
    //-------------------------------------------------------------------------
    // FFT/IFFT Routines
    //-------------------------------------------------------------------------

    //! Computes the FFT of the terms in the 1D array 'arr'. Length of 'arr' is N.
    //! Returns the array of FFT coefficients
    fft_coeff_t* fft_1d(const fft_coeff_t* arr, const unsigned int N);

    //! Computes the FFT of the terms in the 1D array 'arr'. Length of 'arr' is N.
    //! Returns the array of FFT coefficients
    fft_coeff_t* ifft_1d(const fft_coeff_t* arr, const unsigned int N);

    //! Computes the FFT of the terms in the 2D array 'arr'.
    //! Modifies the input array
    void fft_2d(fft_arr_t &arr);

    //! Computes the inverse FFT of the terms in the 2D array 'arr'.
    //! Modifies the input array
    void ifft_2d(fft_arr_t &arr);

    //! Generates the image of the power spectrum of the fourier coefficients in 'arr'
    Image& power_spectrum_image(const fft_arr_t &arr);


    //----------------------------------------------------------------------------
    // Frequency Domain Filters
    //----------------------------------------------------------------------------

    //! Applies the Ideal Lowpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void lowpass(fft_arr_t &arr, double Do); 

    //! Applies the Ideal Highpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void highpass(fft_arr_t &arr, double Do); 

    //! Applies the Butterworth Lowpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void butterworth_lowpass(fft_arr_t &arr, double Do, double n); 

    //! Applies the Butterworth Highpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void butterworth_highpass(fft_arr_t &arr, double Do, double n); 

    //! Applies the Gaussian Lowpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void gaussian_lowpass(fft_arr_t &arr, double Do); 

    //! Applies the Gaussian Highpass filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void gaussian_highpass(fft_arr_t &arr, double Do);

    //! Applies the Homomorphic Gaussian filter to the image
    //! Image is in the frequency domain and the filter modifies the image
    void homomorphic_gaussian(fft_arr_t &arr, double gammaL, double gammaH);

  };



}

#endif
